//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_HPP
#define MESH_HPP

/**
 * \file Mesh.hpp
 *
 * This files contains the definition of a mesh for the Process API
 */

#include <Config.hpp>

#include <cuda/CudaExport.hpp>
#include <EdgeData.hpp>
#include <Geometry.hpp>
#include <Stack.hpp>
#include <SymmetricTensor.hpp>
#include <TransferFunction.hpp>
#include <UnorderedMap.hpp>
#include <UnorderedSet.hpp>
#include <VertexData.hpp>
#include <VVGraph.hpp>

#include <map>
#include <omp.h>
#include <QImage>
#include <QString>
#include <set>
#include <string.h>
#include <typeinfo>

namespace mgx {

  /**
   * Type of the VV graph holding the actual mesh
   */
  typedef graph::VVGraph<VertexData, EdgeData> vvgraph;
  
  /// Type of a vertex
  typedef vvgraph::vertex_t vertex;
  
  /// Type of an edge
  typedef vvgraph::edge_t edge;
  
  typedef unsigned short ushort;
  typedef unsigned int uint;
  typedef unsigned long ulong;
  
  typedef util::Color<float> Colorf;
  
  typedef util::Vector<3, bool> Point3b;
  
  /// Map of an integer to a 3D point
  typedef std::map<int, Point3f> IntPoint3fMap;
  typedef std::unordered_map<int, Point3f> IntPoint3fUMap;
  /// Element in IntPoint3fMap
  typedef std::pair<int, Point3f> IntPoint3fPair;
  /// Map of an integer to a float
  typedef std::map<int, float> IntFloatMap;
  typedef std::unordered_map<int, float> IntFloatUMap;
  /// Element in IntFloatMap
  typedef std::pair<int, float> IntFloatPair;
  /// Map of an integer to another one
  typedef std::map<int, int> IntIntMap;
  typedef std::unordered_map<int, int> IntIntUMap;
  /// Element in IntIntMap
  typedef std::pair<int, int> IntIntPair;
  /// Map of an integer to a host vector of 3 unsigned integers
  typedef std::map<int, HVec3U> IntHVec3uMap;
  typedef std::unordered_map<int, HVec3U> IntHVec3uUMap;
  /// Element in IntHVec3uMap
  typedef std::pair<int, HVec3U> IntHVec3uPair;
  /// Set of vertex ids
  typedef std::set<ulong> VIdSet;
  typedef std::unordered_set<ulong> VIdUSet;
  /// Set of vertices
  typedef std::set<vertex> VtxSet;
  typedef std::unordered_set<vertex> VtxUSet;
  /// Vector of vertices
  typedef std::vector<vertex> VtxVec;
  /// Vector of vvgraphs
  typedef std::vector<vvgraph> VVGraphVec;
  /// Integer, VVGraph pair
  typedef std::pair<int, vvgraph> IntVVGraphPair;
  /// Map of an integer to a vvgraph
  typedef std::map<int, vvgraph> IntVVGraphMap;
  typedef std::unordered_map<int, vvgraph> IntVVGraphUMap;
  /// Map of an integer to a set of vertex ids
  typedef std::map<int, VIdSet> IntVIdSetMap;
  typedef std::unordered_map<int, VIdSet> IntVIdSetUMap;
  /// Element in IntVIdSetMap
  typedef std::pair<int, VIdSet> IntVIdSetPair;
  /// Set of integers
  typedef std::set<int> IntSet;
  typedef std::unordered_set<int> IntUSet;
  /// Vector of integers
  typedef std::vector<int> IntVec;
  /// Vector of floats
  typedef std::vector<float> FloatVec;
  /// Element in IntIntSetMap
  typedef std::pair<int, IntSet> IntIntSetPair;
  /// Map of an integer to a set of integers
  typedef std::map<int, IntSet> IntIntSetMap;
  typedef std::unordered_map<int, IntSet> IntIntSetUMap;
  /// Map of a vertex to an integer
  typedef std::map<vertex, int> VIntMap;
  typedef std::unordered_map<vertex, int> VIntUMap;
  /// Element in IntIntVIdSetMap
  typedef std::pair<IntIntPair, VIdSet> IntIntVIdSetPair;
  /// Element in IntIntFloatMap
  typedef std::pair<IntIntPair, float> IntIntFloatPair;
  /// Map of a pair of integers to a float
  typedef std::map<IntIntPair, float> IntIntFloatMap;
  typedef std::unordered_map<IntIntPair, float> IntIntFloatUMap;
  /// Map of a pair of integers to a set of vertex ids
  typedef std::map<IntIntPair, VIdSet> IntIntVIdSetMap;
  typedef std::unordered_map<IntIntPair, VIdSet> IntIntVIdSetUMap;
  /// Map of an integer to a 3x3 matrix
  typedef std::map<int, Matrix3f> IntMatrix3x3fMap;
  typedef std::unordered_map<int, Matrix3f> IntMatrix3x3fUMap;
  /// Element in IntMatrix3x3fMap
  typedef std::pair<int, Matrix3f> IntMatrix3x3fPair;
  /// Map an integer to a symmetric tensor
  typedef std::map<int, util::SymmetricTensor> IntSymTensorMap;
  typedef std::unordered_map<int, util::SymmetricTensor> IntSymTensorUMap;
  /// Element in IntSymTensorMap
  typedef std::pair<int, util::SymmetricTensor> IntSymTensorPair;
  /// Map of vertex to vertex
  typedef std::map<vertex, vertex> VtxVtxMap;
  typedef std::unordered_map<vertex, vertex> VtxVtxUMap;
}

namespace std {
#ifdef HASH_NEED_TR1
namespace tr1 {
#endif

template <typename T1, typename T2> struct hash<std::pair<T1, T2> > {
  size_t operator()(const std::pair<T1, T2>& v) const
  {
    std::hash<T1> h1;
    std::hash<T2> h2;
    return h1(v.first) ^ h2(v.second);
  }
};

#ifdef HASH_NEED_TR1
} // namspace tr1
#endif
} // namespace std

namespace mgx {

  namespace process {
    class SetupProcess;
  }
  
  /**
   * \class Mesh Mesh.hpp <Mesh.hpp>
   *
   * This class holds the actual mesh as a VV Graph and all sort of properties
   * for it, including visualization ones.
   *
   * Note that, in MorphoGraphX, the user sees the surface, or the mesh. The
   * surface is continuous, when the mesh is the actual connectivity of the
   * vertices.
   */
  class mgx_EXPORT Mesh {
    friend class process::SetupProcess;
  
  public:
    /**
     * Enumeration of the possible visualisations for the surface
     */
    enum SurfView {
      NORMAL,     ///< Show the color option
      LABEL,      ///< Show the label of the triangles
      HEAT,       ///< Show the current heat map
      PARENT      ///< Show the parent label
    };
  
    /**
     * Enumeration of the possible visualisations for the Mesh
     */
    enum MeshView {
      ALL_MESH,         ///< Show the whole mesh
      BORDER_MESH,      ///< Show only the mesh border
      CELL_MESH,        ///< Show only the cell borders (i.e. interface between labels)
      SELECTED_MESH     ///< Show only the selected part of the mesh
    };
  
    /**
     * Enumeration of the coloring modes possible to display a normal surface
     */
    enum Color {
      SIGNAL,      ///< Color using the signal stored on the vertices
      TEXTURE,     ///< Color using the 3D texture of the active stack
      IMAGE        ///< Color using the texture stored with the mesh
    };
  
    /**
     * Create an empty mesh attached to the stack
     */
    Mesh(const Stack* stack);
  
    /**
     * Create an empty mesh attached to the stack
     */
    Mesh(int id, const Stack* stack);
  
    /**
     * Copy the mesh
     */
    Mesh(const Mesh& copy);
  
    /**
     * Mesh destructor
     */
    ~Mesh();
  
    //@{
    ///\name Global properties
    /// Properties for both mesh and surface
  
    /**
     * Id of the current mesh
     *
     * This is the same id used in the Process:mesh(int) method.
     */
    int id() const {
      return _id;
    }
  
    /**
     * Change the id of the mesh.
     *
     * Please do not use this method for meshes attached to a process.
     */
    void setId(int i) {
      _id = i;
    }
  
    /**
     * Id as seen by the user
     *
     * This is, typically id()+1
     */
    int userId() const {
      return _id + 1;
    }
  
    /**
     * Returns the stack associated to this mesh
     */
    const Stack* stack() const {
      return _stack;
    }
  
    // For now, a mesh is always with the same stack
    void setStack(const Stack* s);
  
    /**
     * Returns the name of the mesh file
     */
    const QString& file() const {
      return _file;
    }

    /**
     * Returns the name of the parent label file
     */
    const QString& parentFile() const {
      return _parentFile;
    }

    /**
     * Set the name of the mesh file.
     *
     * This is called when the mesh is loaded or saved. Processes that create
     * a new mesh should call this without a parameter to reset the filename to null.
     */
    void setFile(const QString& file = QString());

     /**
     * Set the name of the parent label file.
     *
     * This is called when the parent labels are loaded or saved. 
     */
    void setParentFile(const QString& file = QString());

    /**
     * Returns true if the mesh is scaled
     */
    bool scaled() const {
      return _scaled;
    }

    /**
     * Set if the mesh is scaled
     */
    void setScaled(bool on = true) {
      _scaled = on;
    }
  
    /**
     * Returns true if the mesh is transformed
     */
    bool transformed() const {
      return _transformed;
    }
    /**
     * Set if the mesh is transformed
     */
    void setTransformed(bool on = true) {
      _transformed = on;
    }
  
    /**
     * Returns true if the bounding box is shown
     */
    bool showBBox() const {
      return _showBBox;
    }
  
    /**
     * Set if the bounding box is shown
     */
    void setShowBBox(bool on = true) {
      _showBBox = on;
    }
  
    /**
     * Returns the current label, without modifying it
     */
    int viewLabel() const {
      return _currLabel;
    }
    /**
     * Sets the current label
     */
    void setLabel(int l)
    {
      changed_label = true;
      _currLabel = l;
    }
    /**
     * Increment the current label and return
     */
    int nextLabel()
    {
      changed_label = true;
      return ++_currLabel;
    }
    /**
     * Returns true if the curr label has been changed during this process
     */
    bool labelChanged() const {
      return changed_label;
    }
  
    /**
     * Reference on the map from each label to a label in a different mesh.
     */
    IntIntMap& parentLabelMap() {
      return _parentLabel;
    }
    /**
     * Const reference on the map from each label to a label in a different mesh.
     */
    const IntIntMap& parentLabelMap() const {
      return _parentLabel;
    }
    /**
     * Returns true if the mesh has a cell structure
     *
     * This means that some vertexes correspond to cell centers and other the
     * junctions.
     */
    bool cells() const {
      return _cells;
    }
    /**
     * Set if the mesh has a cell structure
     *
     * This means that some vertexes correspond to cell centers and other the
     * junctions.
     */
    void setCells(bool value = true) {
      _cells = value;
    }
  
    /**
     * Reset the whole mesh and its properties.
     *
     * In the end, the mesh is empty, so doesn't contain cells, texture and is neither scaled nor transformed.
     *
     * The next label stays untouched, as it is important to keep coordination between different meshes/stacks.
     */
    void reset();
  
    //@}
  
    //@{
    ///\name Structure
  
    /**
     * Get the VV graph representing the topology of the mesh
     */
    vvgraph& graph() {
      return S;
    }
    /**
     * Get the VV graph representing the topology of the mesh
     */
    const vvgraph& graph() const {
      return S;
    }
    /**
     * True if the mesh is empty (i.e. the graph is empty)
     */
    bool empty() const {
      return S.empty();
    }
    /**
     * Returns the number of vertices in the mesh
     */
    size_t size() const {
      return S.size();
    }
  
    /**
     * Returns the label of the triangle (v1,v2,v3)
     */
    int getLabel(vertex v1, vertex v2, vertex v3, bool useParentLabel = true) const;
  
    /**
     * Returns true if (v,n,m) correspond to a canonical triangle.
     *
     * This means it returns true only for one of all the possible triangles
     * made from v, n and m (i.e. (v,n,m), (v,m,n), ...)
     */
    bool uniqueTri(const vertex& v, const vertex& n, const vertex& m) const
    {
      if(v <= n or v <= m or n == m or !S.edge(v, n) or !S.edge(v, m) or !S.edge(n, m))
        return false;
      else
        return true;
    }
  
    /**
     * Update the normal of the vertex \c v in the mesh
     *
     * Returns false if the normal cannot be computed
     */
    bool setNormal(vertex v) {
      return setNormal(S, v);
    }
  
    /**
     * Update the normal of the vertex \c v in \c S
     *
     * Returns false if the normal cannot be computed
     */
    static bool setNormal(vvgraph& S, vertex v);
  
    /**
     * Update all the normals of the mesh
     *
     * Returns false if any normal cannot be computed
     */
    bool setNormals();
  
    /**
     * Reference on the map from label to a position.
     *
     * This position will be intepreted as the center of the label, and used to
     * draw the label number on the mesh or growth directions on cellular mesh.
     *
     */
    IntPoint3fMap& labelCenter() {
      return _labelCenter;
    }
    /**
     * Returns the map from label to position
     */
    const IntPoint3fMap& labelCenter() const {
      return _labelCenter;
    }
  
    /**
     * Reference on the map from label to a normal.
     *
     * This vector will be intepreted as the normal of the label, and used to
     * draw the cell axis (e.g. principal directions of growth, MT orientation...).
     *
     */
    IntPoint3fMap& labelNormal() {
      return _labelNormal;
    }
    /**
     * Returns the map from label to normal
     */
    const IntPoint3fMap& labelNormal() const {
      return _labelNormal;
    }
    /**
     * Reference on the map from label to a position.
     *
     * This position will be intepreted as the center of the label, and used to
     * draw growth directions on a non-cellular mesh.
     *
     */
    IntPoint3fMap& labelCenterVis() {
      return _labelCenterVis;
    }
    /**
     * Returns the map from label to position
     */
    const IntPoint3fMap& labelCenterVis() const {
      return _labelCenterVis;
    }
  
    /**
     * Reference on the map from label to a normal.
     *
     * This vector will be intepreted as the normal of the label, and used to
     * draw the cell axis (e.g. principal directions of growth, MT orientation...).
     *
     */
    IntPoint3fMap& labelNormalVis() {
      return _labelNormalVis;
    }
    /**
     * Returns the map from label to position
     */
    const IntPoint3fMap& labelNormalVis() const {
      return _labelNormalVis;
    }
  
    /**
     * Reference on the map from parent label to a position.
     *
     * This position will be intepreted as the center of the parent label, and used to
     * draw the parent label number on the mesh or growth directions on cellular mesh.
     *
     */
    IntPoint3fMap& parentCenter() {
      return _parentCenter;
    }
    /**
     * Returns the map from parent label to position
     */
    const IntPoint3fMap& parentCenter() const {
      return _parentCenter;
    }
  
    /**
     * Reference on the map from parent label to a normal.
     *
     * This vector will be intepreted as the normal of the parent label, and used to
     * draw the cell axis (e.g. principal directions of growth, MT orientation...).
     *
     */
    IntPoint3fMap& parentNormal() {
      return _parentNormal;
    }
    /**
     * Returns the map from parent label to normal
     */
    const IntPoint3fMap& parentNormal() const {
      return _parentNormal;
    }
    /**
     * Reference on the map from parent label to a position.
     *
     * This position will be intepreted as the center of the parent label, and used to
     * draw growth directions on a non-cellular mesh.
     *
     */
    IntPoint3fMap& parentCenterVis() {
      return _parentCenterVis;
    }
    /**
     * Returns the map from parent label to position
     */
    const IntPoint3fMap& parentCenterVis() const {
      return _parentCenterVis;
    }
  
    /**
     * Reference on the map from parent label to a normal.
     *
     * This vector will be intepreted as the normal of the parent label, and used to
     * draw the cell axis (e.g. principal directions of growth, MT orientation...).
     *
     */
    IntPoint3fMap& parentNormalVis() {
      return _parentNormalVis;
    }
    /**
     * Returns the map from parent label to position
     */
    const IntPoint3fMap& parentNormalVis() const {
      return _parentNormalVis;
    }
  
    /**
     * Reference on the map from labels to the set of neighbor labels.
     *
     * A label is neighbors of another ones if they share at least an edge of the mesh.
     *
     * \note You shouldn't need to modify this structure directly. You should
     * instead call Mesh::updateWallGeometry()
     */
    IntIntSetMap& labelNeighbors() {
      return _labelNeighbors;
    }
    /**
     * Returns the map from labels to the set of neighbors labels.
     */
    const IntIntSetMap& labelNeighbors() const {
      return _labelNeighbors;
    }
    /**
     * Reference on the map from each wall to the list of vertexes in the wall.
     *
     * A wall is identified as the interface between two labels, and as such, is given as a pair of labels. Note that
     * there is a difference between the wall (i,j) and (j,i), as the first label is considered as the interior of the
     * wall, and used as such when calculating the signal along a wall.
     *
     * \note You shouldn't need to modify this structure directly. You should
     * instead call Mesh::updateWallGeometry()
     */
    IntIntVIdSetMap& wallVId() {
      return _wallVId;
    }
    /**
     * Returns the map from each wall to the list of vertexes in the wall.
     */
    const IntIntVIdSetMap& wallVId() const {
      return _wallVId;
    }
    /**
     * Reference on the map from each wall to the length of the wall.
     *
     * The length is calculated as the sum of the edges making the wall. This means the number returned is will depend
     * on the discretization of the surface.
     *
     * \note You shouldn't need to modify this structure directly. You should
     * instead call Mesh::updateWallGeometry()
     */
    IntIntFloatMap& wallGeom() {
      return _wallGeom;
    }
    /**
     * Reference on the map from each wall to the length of the wall.
     */
    const IntIntFloatMap& wallGeom() const {
      return _wallGeom;
    }
  
    /**
     * Update labelNeighbors, wallVId and wallGeom to reflect current data structure
     * \param borderSize Distance from the wall up to which the triangles are considered part of the wall. This is
     * important mostly for signal intensity.
     */
    void updateWallGeometry(float borderSize);
  
    /**
     * True if (v,n,m) is on the border of the \c label.
     *
     * \c rp will be set to the wall the triangle is on, if any.
     */
    bool isBordTriangle(int label, vertex v, vertex n, vertex m, IntIntPair& rp) const;
  
    /**
     * Mark all triangles closer than borderSize from a border as being
     * a border triangle
     */
    void markBorder(float borderSize);
  
    /**
     * Search for the vertex closest to position of cell center
     * fills in labelCenter, labelNormal, labelCenterVis, labelNormalVis
     */
    void updateCentersNormals();
  
    /**
     * Prepare the axis drawing structures
     */
    void prepareAxisDrawing(float scaleAxisLength, Colorf colorPositive, Colorf colorNegative,
                            const Point3b& showAxis = Point3b(true, true, true));   // ,float DisplayThreshold)
  
    //@}
  
    //@{
    ///\name Surface methods
    /**
     * Return the transfer function used to draw the surface in normal mode
     */
    TransferFunction surfFct() const {
      return _surf_fct;
    }
    /**
     * Change the transfer function used to draw the surface in normal mode
     */
    void setSurfFct(const TransferFunction& f)
    {
      if(_surf_fct != f) {
        _surf_fct = f;
        changed_surf_function = true;
      }
    }
    /**
     * Returns true if the normal transfer function has been changed during the
     * current process.
     */
    bool surfFctChanged() const {
      return changed_surf_function;
    }
  
    /**
     * Return the transfer function used to draw the surface in heat mode
     */
    TransferFunction heatFct() const {
      return _heat_fct;
    }
    /**
     * Change the transfer function used to draw the surface in heat mode
     */
    void setHeatFct(const TransferFunction& f)
    {
      if(_heat_fct != f) {
        _heat_fct = f;
        changed_heat_function = true;
      }
    }
    /**
     * Returns true if the heat transfer function has been changed during the
     * current process.
     */
    bool heatFctChanged() const {
      return changed_heat_function;
    }
  
    /**
     * Get the current opacity level of the surface
     */
    float opacity() const {
      return _opacity;
    }
    /**
     * Change the current opactity level of the surface
     */
    void setOpacity(float f)
    {
      if(f < 0)
        _opacity = 0;
      else if(f > 1)
        _opacity = 1;
      else
        _opacity = f;
    }
    /**
     * Get the current brightness of the surface
     */
    float brightness() const {
      return _brightness;
    }
    /**
     * Change the current brightness of the surface
     */
    void setBrightness(float f)
    {
      if(f < 0)
        _brightness = 0;
      else if(f > 1)
        _brightness = 1;
      else
        _brightness = f;
    }
  
    /**
     * True if the surface is currently visible to the user
     */
    bool isSurfaceVisible() const {
      return _isSurfaceVisible;
    }
    /**
     * Show the surface to the user
     */
    void showSurface() {
      _isSurfaceVisible = true;
    }
    /**
     * Hide the surface from the user
     */
    void hideSurface() {
      _isSurfaceVisible = false;
    }
  
    /**
     * Show the heat map of the surface, if any
     */
    void showNormal() {
      _show = NORMAL;
    }
    /**
     * Show the labels of the surface triangles
     */
    void showLabel() {
      _show = LABEL;
    }
    /**
     * Show the color of the surface
     */
    void showHeat() {
      _show = HEAT;
    }
    /**
     * Show the color of the surface
     */
    void showParent() {
      _show = PARENT;
    }
  
    /**
     * Change the view mode
     */
    void show(SurfView view) {
      _show = view;
    }
  
    /**
     * Set the current color to be the signal (i.e. intensity) on each triangle
     */
    void showSignal() {
      _color = SIGNAL;
    }
    /**
     * Set the current color to be the 3D texture at the position of the point.
     *
     * This represent the surface as a cut through the volume. The 3D texture
     * selected correspond to the current one for the stack attached to this
     * mesh.
     */
    void showTexture() {
      _color = TEXTURE;
    }
  
    /**
     * Set the current color to be the texture attached to the mesh.
     */
    void showImage() {
      _color = IMAGE;
    }
  
    /**
     * Change the color mode
     */
    void show(Color color) {
      _color = color;
    }
  
    /**
     * Returns if the back surface is culled
     */
    bool culling() const {
      return _culling;
    }
    /**
     * Cull the back surface
     */
    void setCulling(bool cul) {
      _culling = cul;
    }
  
    /**
     * Return if the surface is rendered blended or not
     */
    bool blending() const {
      return _blending;
    }
    /**
     * Set the blending attribute
     */
    void setBlending(bool b) {
      _blending = b;
    }
  
    /**
     * Returns the current visualisation for the surface
     */
    SurfView toShow() const {
      return _show;
    }
    /**
     * Returns the current coloring for the surface
     */
    Color coloring() const {
      return _color;
    }
  
    /**
     * True if the surface has a texture attached to it
     */
    bool hasImgTex() const {
      return _imgtex;
    }
    /**
     * Get the current texture attached to the surface
     */
    const QImage& imgTex() const {
      return _image;
    }
    /**
     * Set the texture attached to the surface
     */
    void setImgTex(const QImage& img)
    {
      _imgtex = true;
      _image = img;
    }
    /**
     * Remove any texture attached to the surface
     */
    void clearImgTex()
    {
      _imgtex = false;
      _image = QImage();
    }
  
    //@}
  
    //@{
    ///\name Mesh methods
    /**
     * Is the mesh currently visible to the user
     */
    bool isMeshVisible() const {
      return _isMeshVisible;
    }
    /**
     * Make the mesh visible to the user
     */
    void showMesh() {
      _isMeshVisible = true;
    }
    /**
     * Hide the mesh from the user
     */
    void hideMesh() {
      _isMeshVisible = false;
    }
  
    /**
     * Return the current view mode for the mesh
     */
    MeshView meshView() const {
      return _meshView;
    }
  
    /**
     * Set the mesh view to all
     */
    void showAllMesh() {
      _meshView = ALL_MESH;
    }
    /**
     * Set the mesh view to cell border only
     */
    void showCellMesh() {
      _meshView = CELL_MESH;
    }
    /**
     * Set the mesh view to mesh border only
     */
    void showBorderMesh() {
      _meshView = BORDER_MESH;
    }
    /**
     * Set the mesh view to selected mesh only
     */
    void showSelectedMesh() {
      _meshView = SELECTED_MESH;
    }
  
    /**
     * Change the mesh view mode
     */
    void show(MeshView view) {
      _meshView = view;
    }
  
    /**
     * Are mesh lines visible
     */
    bool showMeshLines() const {
      return _showMeshLines;
    }
    /**
     * Are mesh points visible
     */
    bool showMeshPoints() const {
      return _showMeshPoints;
    }
    /**
     * Are cell numbers visible
     */
    bool showMeshCellMap() const {
      return _showMeshCellMap;
    }
  
    /**
     * Set if the mesh visible at all
     */
    void setShowMesh(bool show) {
      _isMeshVisible = show;
    }
    /**
     * Set if the mesh lines are visible
     */
    void setShowMeshLines(bool show) {
      _showMeshLines = show;
    }
    /**
     * Set if the mesh points are visible
     */
    void setShowMeshPoints(bool show) {
      _showMeshPoints = show;
    }
    /**
     * Set if the cell numbers are visible
     */
    void setShowMeshCellMap(bool show) {
      _showMeshCellMap = show;
    }
  
    /**
     * Returns a vector with the list of selected vertices.
     */
    std::vector<vertex> selectedVertices() const;
  
    /**
     * Returns a vector with the list of active vertices.
     *
     * These are either the selected vertices, or all of them if no vertex was selected.
     */
    std::vector<vertex> activeVertices() const;
  
    // These next two should really be algorithms for the graph as they don't rely on
    // any Mesh variables.
    /**
     * Starting with a set of vertices, extend to connected region
     */
    void getConnectedVertices(const vvgraph& S, vvgraph& vSet) const;
  
    /**
     * Get a list of connected regions (3D cells)
     */
    int getConnectedRegions(const vvgraph& S, VVGraphVec& cellVId, VIntMap& vtxCell) const;
  
    /**
     * Get the neighborhood of a vertex
     */
    static void getNhbd(const vvgraph& S, const vertex& v, float radius, VtxSet& nbs);
  
    /**
     * Correct selection
     *
     * Make sure the vertices are "correctly" selected. This means that:
     *  * border vertices will be selected if all their non-border neighbors are
     *  * border vertices will be selected if they don't have a non-border neighbors and at least one border neighbor is
     *  selected
     *  * if the selection is inclusive, border vertices will be selected if at least one non-border neighbor is
     *  selected
     */
    void correctSelection(bool inclusive);
  
    //@}
  
    //@{
    ///\name Updating the graphical representation
    /**
     * To be called when the mesh structure changed
     */
    void updateAll();
    /**
     * To be called if properties of the triangles changed
     */
    void updateTriangles();
    /**
     * To be called if properties of the lines changed
     */
    void updateLines();
    /**
     * To be called if the position of the points changed
     */
    void updatePositions();
    /**
     * To be called if the selection of the points changed
     */
    void updateSelection();
    /** Returns what changed in the current process
     */
    int changes() const {
      return _changed;
    }
    //@}
  
    //@{
    /**
     * \name Heat Map methods
     * The heat map associate a heat (from 0 to 1) to a label or a wall.
     *
     * As the heat is always from 0 to 1, the bounds of the heat map have to be
     * specified separatly for presentation purposes and correspond to the
     * actual value associated to 0 and 1.
     */
    /** Reference on the map from label to intensity
     */
    IntFloatMap& labelHeat() {
      return _labelHeat;
    }
    /** Returns the map from label to intensity
     */
    const IntFloatMap& labelHeat() const {
      return _labelHeat;
    }
    /** Reference on the map from wall (i.e. pair of vertex id) to intensity
     */
    IntIntFloatMap& wallHeat() {
      return _wallHeat;
    }
    /** Returns the map from wall (i.e. pair of vertex id) to intensity
     */
    const IntIntFloatMap& wallHeat() const {
      return _wallHeat;
    }
    /**
     * Reference on the unit used for the current heat map.
     *
     * It is shown on the color bar.
     */
    QString& heatMapUnit() {
      return _heatMapUnit;
    }
    /// Returns the unit used for the current heat map
    const QString& heatMapUnit() const {
      return _heatMapUnit;
    }
    /// Reference on the upper and lower bounds for the heat.
    Point2f& heatMapBounds() {
      return _heatMapBounds;
    }
    /// Return the upper and lower bounds for the heat.
    const Point2f& heatMapBounds() const {
      return _heatMapBounds;
    }
  
    /**
     * Return the mapping of tensors for each cell
     *
     * The tensors are encoded as three vectors: ev1, ev2, evals.
     * Where ev1 and ev2 are the two first eigenvectors and evals is the vector with the three eigenvalues.
     */
    IntSymTensorMap& cellAxis() {
      return _cellAxis;
    }
  
    /**
     * Returns the map from label to cells axis (array of 3 Point3f), scaled for visualisation
     */
    IntMatrix3x3fMap& cellAxisScaled() {
      return _cellAxisScaled;
    }                                                                // added by ALR
  
    /**
     * Returns the map from label to cells axis colors (array of 3 Point3f)
     */
    IntMatrix3x3fMap& cellAxisColor() {
      return _cellAxisColor;
    }                                                              // added by ALR
  
    /** Returns the cell axis width
     */
    float& cellAxisWidth() {
      return _cellAxisWidth;
    }                                                   // added by ALR
    /** Returns a reference to the cell axis z-offset for drawing
     */
    float& cellAxisOffset() {
      return _cellAxisOffset;
    }
    /**
     * Return the type of the current cell axis.
     *
     * If an empty string, there is no cell axis.
     */
    const QString& cellAxisType() const {
      return _cellAxisType;
    }
    /**
     * Return the type of the current cell axis.
     *
     * If an empty string, there is no cell axis.
     */
    QString& cellAxisType() {
      return _cellAxisType;
    }
    /**
     * Unit for the length of the cell axis
     */
    const QString& cellAxisUnit() const {
      return _cellAxisUnit;
    }
    /**
     * Unit for the length of the cell axis
     */
    QString& cellAxisUnit() {
      return _cellAxisUnit;
    }
    /**
     * Remove any cell axis
     */
    void clearCellAxis();
    /** Returns a reference to the map of correspondence between vertices in mesh1 and mesh2 (store vertices positions)
     */
    IntMatrix3x3fMap& vvCorrespondance() {
      return _VVCorrespondance;
    }                                                                    // added by ALR
    /** Returns a boolean for visualization of lines between corresponding vertices
     */
    bool& showVVCorrespondance() {
      return _showVVCorrespondance;
    }                                                                // added by ALR
  
    //@}
  
    //@{
    /**
     * \name Signal methods
     * The signal spans from 0 to 1, but may correspond to any range of values.
     * These methods specifies the range and the units.
     */
    /**
     * Reference on the unit used for the signal.
     *
     * It is shown on the color bar.
     */
    QString& signalUnit() {
      return _signalUnit;
    }
    /// Returns the unit used for the signal
    const QString& signalUnit() const {
      return _signalUnit;
    }
    /// Reference on the upper and lower bounds for the signal.
    Point2f& signalBounds() {
      return _signalBounds;
    }
    /// Return the upper and lower bounds for the signal.
    const Point2f& signalBounds() const {
      return _signalBounds;
    }
    //@}
  
    bool needsColorbar() const
    {
      return (isSurfaceVisible() and ((toShow() == HEAT and (_labelHeat.size() > 1 or _wallHeat.size() > 1))
                                      or (toShow() == NORMAL and coloring() == SIGNAL)));
    }
  
    const BoundingBox3f& boundingBox() const {
      return _bbox;
    }
    BoundingBox3f& boundingBox() {
      return _bbox;
    }
    void setBoundingBox(const BoundingBox3f& bbox) {
      _bbox = bbox;
    }
    void updateBBox();
  
  private:
    void resetModified();
  
    vvgraph S;
    TransferFunction _surf_fct, _heat_fct;
    IntFloatMap _labelHeat;
    IntIntFloatMap _wallHeat;
    IntPoint3fMap _labelCenter;
    IntPoint3fMap _labelCenterVis;
    IntPoint3fMap _labelNormal;
    IntPoint3fMap _labelNormalVis;
    IntPoint3fMap _parentCenter;
    IntPoint3fMap _parentCenterVis;
    IntPoint3fMap _parentNormal;
    IntPoint3fMap _parentNormalVis;
    IntSymTensorMap _cellAxis;
    IntMatrix3x3fMap _cellAxisScaled;
    IntMatrix3x3fMap _cellAxisColor;
    float _cellAxisWidth;
    float _cellAxisOffset;
    QString _cellAxisType;
    QString _cellAxisUnit;
    IntMatrix3x3fMap _VVCorrespondance;
    bool _showVVCorrespondance;
    IntIntSetMap _labelNeighbors;
    IntIntVIdSetMap _wallVId;
    IntIntFloatMap _wallGeom;
    QString _signalUnit, _heatMapUnit;
    Point2f _signalBounds, _heatMapBounds;
    bool changed_surf_function, changed_heat_function;
    bool changed_label;
    int _changed;
    SurfView _show;
    Color _color;
    bool _culling, _blending;
    bool _isSurfaceVisible, _isMeshVisible;
    int _currLabel;
    IntIntMap _parentLabel;
    float _opacity, _brightness;
    MeshView _meshView;
    bool _showMeshLines, _showMeshPoints, _showMeshCellMap;
    bool _cells, _imgtex;
    QString _file;
    QString _parentFile;
    bool _scaled;
    bool _transformed;
    bool _showBBox;
    int _id;
    const Stack* _stack;
    QImage _image;
  
    BoundingBox3f _bbox;
  
    void init();
  };
}

#endif 
