//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef ITKFILTER_HPP
#define ITKFILTER_HPP

#include <ITKConfig.hpp>
#include <ITKProcess.hpp>

namespace mgx {
namespace process {
class mgxITK_EXPORT ITKSmoothingRecursiveGaussianImageFilter : public StackProcess {
public:
  ITKSmoothingRecursiveGaussianImageFilter(const StackProcess& process)
    : Process(process), StackProcess(process) {}

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_EMPTY))
      return false;
    Stack* s = currentStack();
    Store* input = s->currentStore();
    Store* output = s->work();
    bool res = (*this)(input, output, parms[0].toFloat());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output, float radius);

  QString name() const {
    return "ITK Smoothing Recursive Gaussian Image Filter";
  }
  QString folder() const {
    return "ITK/Filters";
  }
  QString description() const {
    return "Smoothing Recursive Gaussian Blur";
  }
  QStringList parmNames() const {
    return QStringList() << QString("Radius (%1)").arg(UM);
  }
  QStringList parmDescs() const {
    return QStringList() << QString("Radius (%1)").arg(UM);
  }
  QStringList parmDefaults() const {
    return QStringList() << "1.0";
  }
  QIcon icon() const {
    return QIcon(":/images/Blur.png");
  }
};

class ITKCurvatureFlowImageFilter : public StackProcess {
public:
  ITKCurvatureFlowImageFilter(const StackProcess& process)
    : Process(process), StackProcess(process) {}

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_EMPTY))
      return false;
    Stack* s = currentStack();
    Store* input = s->currentStore();
    Store* output = s->work();
    bool res = (*this)(input, output, parms[0].toFloat(), parms[1].toInt());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output, float timestep, int steps);

  QString name() const {
    return "ITK Curvature Flow Image Filter";
  }
  QString folder() const {
    return "ITK/Filters";
  }
  QString description() const {
    return "Curvature Flow Edge Preserving Smoothing";
  }
  QStringList parmNames() const
  {
    return QStringList() << "TimeStep"
                         << "Steps";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "TimeStep"
                         << "Steps";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << ".0.0625"
                         << "10";
  }
  QIcon icon() const {
    return QIcon(":/images/Blur.png");
  }
};

class ITKCurvatureAnisotropicDiffusionImageFilter : public StackProcess {
public:
  ITKCurvatureAnisotropicDiffusionImageFilter(const StackProcess& process)
    : Process(process), StackProcess(process) {}

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_EMPTY))
      return false;
    Stack* s = currentStack();
    Store* input = s->currentStore();
    Store* output = s->work();
    bool res = (*this)(input, output, parms[0].toFloat(), parms[1].toInt(), parms[2].toFloat());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output, float timestep, int iterations, float conductance);

  QString name() const {
    return "ITK Curvature Anisotropic Diffusion Image Filter";
  }
  QString folder() const {
    return "ITK/Filters";
  }
  QString description() const {
    return "Curvature Anisotropic Diffusion";
  }
  QStringList parmNames() const
  {
    return QStringList() << "TimeStep"
                         << "Iterations"
                         << "Conductance";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "TimeStep"
                         << "Iterations"
                         << "Conductance";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "0.03"
                         << "5"
                         << "9.0";
  }
  QIcon icon() const {
    return QIcon(":/images/Blur.png");
  }
};

class ITKGradientMagnitudeRecursiveGaussianImageFilter : public StackProcess {
public:
  ITKGradientMagnitudeRecursiveGaussianImageFilter(const StackProcess& process)
    : Process(process), StackProcess(process) {}

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_EMPTY))
      return false;
    Stack* s = currentStack();
    Store* input = s->currentStore();
    Store* output = s->work();
    bool res = (*this)(input, output, parms[0].toFloat());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output, float sigma);

  QString name() const {
    return "ITK Gradient Magnitude Recursive Gaussian Image Filter";
  }
  QString folder() const {
    return "ITK/Filters";
  }
  QString description() const {
    return "Gradient Magnitude Recursive Gaussian";
  }
  QStringList parmNames() const {
    return QStringList() << QString("Sigma (%1)").arg(UM);
  }
  QStringList parmDescs() const {
    return QStringList() << QString("Sigma (%1)").arg(UM);
  }
  QStringList parmDefaults() const {
    return QStringList() << "0.5";
  }
  QIcon icon() const {
    return QIcon(":/images/Blur.png");
  }
};

class ITKSigmoidImageFilter : public StackProcess {
public:
  ITKSigmoidImageFilter(const StackProcess& process)
    : Process(process), StackProcess(process) {}

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_EMPTY))
      return false;
    Stack* s = currentStack();
    Store* input = s->currentStore();
    Store* output = s->work();
    bool res = (*this)(input, output, parms[0].toFloat(), parms[1].toFloat());
    if(res) {
      input->hide();
      output->show();
    }
    return res;
  }

  bool operator()(const Store* input, Store* output, float alpha, float beta);

  QString name() const {
    return "ITK Sigmoid Image Filter";
  }
  QString folder() const {
    return "ITK/Filters";
  }
  QString description() const {
    return "Sigmoid Filter";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Alpha"
                         << "Beta";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Alpha"
                         << "Beta";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "1000.0"
                         << "2000.0";
  }
  QIcon icon() const {
    return QIcon(":/images/Blur.png");
  }
};
}
}

#endif
