//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESS_CREATION_HPP
#define MESH_PROCESS_CREATION_HPP

#include <Process.hpp>

#include <Polygonizer.hpp>

namespace mgx {
namespace process {
///\addtogroup MeshProcess
///@{
/**
 * \class MarchingCubeSurface MeshProcessCreation.hpp <MeshProcessCreation.hpp>
 *
 * Find the surface of a volume defined by an iso-surface on the image
 * intensity using a variation on the Marching Cube method.
 */
class mgxBase_EXPORT MarchingCubeSurface : public MeshProcess, public util::ImplicitFunction {
public:
  MarchingCubeSurface(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
    , progress("Marching Cubes Surface", 0)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_NON_EMPTY).mesh())
      return false;

    _stack = currentStack();
    _store = _stack->currentStore();
    Mesh* mesh = currentMesh();

    bool result = (*this)(mesh, _store, parms[0].toFloat(), parms[1].toInt());
    if(result)
      mesh->showSignal();
    if(result and not mesh->isSurfaceVisible() and not mesh->isMeshVisible())
      mesh->showSurface();
    return result;
  }

  bool operator()(Mesh* mesh, const Store* store, float cubeSize, int threshold);

  QString folder() const {
    return "Creation";
  }
  QString name() const {
    return "Marching Cubes Surface";
  }
  QString description() const
  {
    return "Extract surface mesh with marching cubes. A threshold to 0 will interpret the image as binary.";
  }
  QStringList parmNames() const {
    return QStringList() << QString("Cube size (%1)").arg(UM) << "Threshold";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Size for the marching cubes."
                         << "Minimal signal used for surface extraction.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "5.0"
                         << "5000";
  }
  QIcon icon() const {
    return QIcon(":/images/MarchCubes.png");
  }

  bool eval(Point3f p);

private:
  int _threshold;
  const Stack* _stack;
  const Store* _store;
  Progress progress;
};

/**
 * \class MarchingCubeSurface MeshProcessCreation.hpp <MeshProcessCreation.hpp>
 *
 * Find the surface of the segmented volumes using a variation on
 * the Marching Cube method.
 */
class mgxBase_EXPORT MarchingCube3D : public MeshProcess, public util::ImplicitFunction {
public:
  MarchingCube3D(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
    , progress("Marching Cubes 3D", 0)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().store(STORE_LABEL).mesh())
      return false;

    _stack = currentStack();
    _store = _stack->currentStore();
    Mesh* mesh = currentMesh();

    bool result = (*this)(mesh, _store, parms[0].toFloat(), parms[1].toInt(), parms[2].toInt(), parms[3].toInt());
    if(result)
      mesh->showLabel();
    if(result and not mesh->isSurfaceVisible() and not currentMesh()->isMeshVisible())
      currentMesh()->showSurface();
    return result;
  }

  bool operator()(Mesh* mesh, const Store* store, float cubeSize, int minVoxels, int smooth, int singleLabel);

  QString folder() const {
    return "Creation";
  }
  QString name() const {
    return "Marching Cubes 3D";
  }
  QString description() const {
    return "Extract 3D meshes with marching cubes";
  }
  QStringList parmNames() const
  {
    return QStringList() << QString("Cube size (%1)").arg(UM) << "Min Voxels"
                         << "Smooth Passes"
                         << "Label";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Size for the marching cubes."
                         << "Minimal number of voxels for extracting surface."
                         << "Smooth Passes."
                         << "Extract surface only for this label.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "5.0"
                         << "0"
                         << "3"
                         << "0";
  }
  QIcon icon() const {
    return QIcon(":/images/MarchCubes3D.png");
  }

  bool eval(Point3f p);

private:
  int _label;
  const Stack* _stack;
  const Store* _store;
  Progress progress;
};

/**
 * \class CuttingSurfMesh MeshProcessCreation.hpp <MeshProcessCreation.hpp>
 *
 * Create a surface matching the current cutting surface, whether it's a
 * plane or a Bezier surface.
 */
class mgxBase_EXPORT CuttingSurfMesh : public MeshProcess {
public:
  CuttingSurfMesh(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh())
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* mesh);

  QString folder() const {
    return "Creation";
  }
  QString name() const {
    return "Mesh Cutting Surface";
  }
  QString description() const {
    return "Make mesh from cutting surface";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/MakePlane.png");
  }
};
///@}
} // namespace process
} // namespace mgx

#endif
