//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESS_SEGMENTATION_HPP
#define MESH_PROCESS_SEGMENTATION_HPP

#include <Process.hpp>

class Ui_LoadHeatMap;

namespace mgx {
namespace process {
///\addtogroup MeshProcess
///@{
/**
 * \class SegmentMesh MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Segment the mesh using a seeded watershed algorithm.
 *
 * The mesh is assumed to already contain seeds for the watershed. For proper
 * use on this process, you should read the protocols and manuals published.
 */
class mgxBase_EXPORT SegmentMesh : public MeshProcess {
public:
  SegmentMesh(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh())
      return false;
    return (*this)(currentMesh(), parms[0].toUInt(), currentMesh()->activeVertices());
  }

  bool operator()(Mesh* mesh, uint maxsteps, const std::vector<vertex>& to_segment);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Watershed Segmentation";
  }
  QString description() const {
    return "Propagate labels on mesh using the watershed algorithm.";
  }
  QStringList parmNames() const {
    return QStringList() << "Steps";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Number of propagation steps performed before updating display. "
                            "Increase value to run faster.";
  }
  QStringList parmDefaults() const {
    return QStringList() << "20000";
  }
  QIcon icon() const {
    return QIcon(":/images/SegmentMesh.png");
  }
};

/**
 * \class SegmentClear MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Remove any segmentation from the current mesh.
 */
class mgxBase_EXPORT SegmentClear : public MeshProcess {
public:
  SegmentClear(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList&)
  {
    if(!checkState().mesh())
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* mesh);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Segmentation Clear";
  }
  QString description() const {
    return "Clean all mesh labels.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/SegmentClear.png");
  }
};

/**
 * \class MeshRelabel MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Relabel a segmented mesh.
 *
 * This process will relabel all the mesh. The labels are assigned in an
 * arbitrary, non-random, manner.
 */
class mgxBase_EXPORT MeshRelabel : public MeshProcess {
public:
  MeshRelabel(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_LABEL))
      return false;
    Mesh* m = currentMesh();
    int start = parms[0].toInt();
    int step = parms[1].toInt();
    return (*this)(m, start, step);
  }

  bool operator()(Mesh* m, int start, int step);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Relabel";
  }
  QString description() const {
    return "Relabel a mesh randomly to use consecutive labels.";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Label Start"
                         << "Label Step";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Starting label."
                         << "Increment step.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "1"
                         << "1";
  }
  QIcon icon() const {
    return QIcon(":/images/Relabel.png");
  }
};

/**
 * \class GrabLabelsSegment MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Grab labels from other mesh (3D meshes).
 */
class mgxBase_EXPORT GrabLabelsSegment : public MeshProcess {
public:
  GrabLabelsSegment(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY, 0).mesh(MESH_NON_EMPTY, 1))
      return false;

    Mesh* mesh1, *mesh2;
    if(currentMesh() == mesh(0)) {
      mesh1 = mesh(0);
      mesh2 = mesh(1);
    } else if(currentMesh() == mesh(1)) {
      mesh2 = mesh(0);
      mesh1 = mesh(1);
    } else
      return false;

    bool res = (*this)(mesh1, mesh2, parms[0].toFloat());
    return res;
  }

  bool operator()(Mesh* mesh1, Mesh* mesh2, float tolerance);

  QString name() const {
    return "3D Grab Labels";
  }
  QString description() const {
    return "Grab labels from other mesh (3D meshes).";
  }
  QString folder() const {
    return "Segmentation";
  }
  QStringList parmNames() const {
    return QStringList() << QString("Tolerance (%1)").arg(UM);
  }
  QStringList parmDescs() const {
    return QStringList() << "Maximal distance between matching cells. ";
  }
  QStringList parmDefaults() const {
    return QStringList() << "5.0";
  }
  QIcon icon() const {
    return QIcon(":/images/GrabLabels.png");
  }
};

/**
 * \class LabelSelected MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Change the label of the selected vertices.
 */
class mgxBase_EXPORT LabelSelected : public MeshProcess {
public:
  LabelSelected(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_LABEL))
      return false;
    return (*this)(currentMesh(), parms[0].toInt());
  }

  bool operator()(Mesh* mesh, int label);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Label Selected Vertices";
  }
  QString description() const {
    return "Label selected vertices with the same value, which can be a new label";
  }
  QStringList parmNames() const {
    return QStringList() << "Label";
  }
  QStringList parmDescs() const {
    return QStringList() << "Asign this label to all selected vertices.";
  }
  QStringList parmDefaults() const {
    return QStringList() << "0";
  }
  QIcon icon() const {
    return QIcon(":/images/LabelSelected.png");
  }
};

/**
 * \class RelabelCells3D MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Relabel 3D cells. Each connected region will be given a unique label.
 */
class mgxBase_EXPORT RelabelCells3D : public MeshProcess {
public:
  RelabelCells3D(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    Mesh* mesh = currentMesh();
    bool res = (*this)(mesh, parms[0].toInt(), parms[1].toInt());
    if(res)
      mesh->showLabel();
    return res;
  }

  bool operator()(Mesh* mesh, int label_start, int label_step);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Relabel 3D Cells";
  }
  QString description() const
  {
    return "Relabel 3D cells (connected regions)."
           " A start label of -1 is used to indicate continuing using current labels.";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Label start"
                         << "Label step";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Starting label. Use -1 to continue from last current label."
                         << "Increment step.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "-1"
                         << "1";
  }
  QIcon icon() const {
    return QIcon(":/images/CellFiles3D.png");
  }
};

/**
 * \class MeshCombineRegions MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Combine over-segmented regions, based on mesh signal.
 */
class MeshCombineRegions : public MeshProcess {
public:
  MeshCombineRegions(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return (*this)(currentMesh(), parms[0].toFloat(), parms[1].toFloat());
  }

  bool operator()(Mesh* mesh, float borderDist, float Difference);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Combine Labels";
  }
  QString description() const {
    return "Combine over-segmented regions, based on mesh signal.";
  }
  QStringList parmNames() const {
    return QStringList() << QString("Border Distance (%1)").arg(UM) << "Threshold";
  }
  QStringList parmDescs() const
  {
    return QStringList() << "Half of the cell border width on the mesh."
                         << "Ratio of border signal over internal signal. If a border between 2 cells "
              "has a ratio below the threshold, the cells are fused.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "1"
                         << "1.5";
  }
  QIcon icon() const {
    return QIcon(":/images/Merge.png");
  }
};

/**
 * \class MeshAutoSegment MeshProcessSegmentation.hpp <MeshProcessSegmentation.hpp>
 *
 * Auto-Segmentation of the mesh surface based on signal. Combines blurring, auto-seeding,
 * label propagation by watershed and fusion of over-segmented cells.
 */
class MeshAutoSegment : public MeshProcess {
public:
  MeshAutoSegment(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return (*this)(currentMesh(), stringToBool(parms[0]), stringToBool(parms[1]), parms[2].toFloat(),
                   parms[3].toFloat(), parms[4].toFloat(), parms[5].toFloat(), parms[6].toFloat(),
                   parms[7].toFloat());
  }

  bool operator()(Mesh* mesh, bool updateView, bool normalize, float gaussianRadiusCell, float localMinimaRadius,
                  float gaussianRadiusWall, float normalizeRadius, float borderDist, float threshold);

  QString folder() const {
    return "Segmentation";
  }
  QString name() const {
    return "Auto-Segmentation";
  }
  QString description() const
  {
    return "Auto-Segmentation of the mesh surface based on signal. Combines blurring, auto-seeding, "
           "label propagation by watershed and fusion of over-segmented cells.";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Update"
                         << "Normalize" << QString("Blur Cell Radius (%1)").arg(UM)
                         << QString("Auto-Seed Radius (%1)").arg(UM) << QString("Blur Borders Radius (%1)").arg(UM)
                         << QString("Normalization Radius (%1)").arg(UM) << QString("Border Distance (%1)").arg(UM)
                         << "Combine Threshold";
  }
  QStringList parmDescs() const
  {
    return QStringList()
           << "Option to update mesh drawing while running processes."
           << "Option to normalize mesh signal before merging the cells. Most usefull in case of low contrast mesh "
              "signal."
           << "Radius used for gaussian blur of mesh signal. In normal cases, should be equal to auto-seed radius."
           << "Radius used for auto-seeding the mesh, based on local minima of signal. Should be equal to radius "
              "of smallest cells."
           << "Radius used for gaussian blur of signal on cell borders before watershed segmentation. Use small "
              "values (roughly, half the border width) to avoid border signal distortion."
           << "Radius used for local normalization of signal. Roughly, radius of largest cells."
           << "Half of the cell border width after blurring. Used for fusion of over-segmented cells."
           << "Ratio of border signal over internal signal. If a borders between 2 cells have a ratio below the "
              "threshold, the cells are fused.";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "Yes"
                         << "Yes"
                         << "2.0"
                         << "2.0"
                         << "1"
                         << "5.0"
                         << "1.0"
                         << "1.8";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = booleanChoice();
    map[1] = booleanChoice();
    return map;
  }
  QIcon icon() const {
    return QIcon(":/images/SegmentMesh.png");
  }
};
///@}
} // namespace process
} // namespace mgx

#endif
