//
// This file is part of MorphoGraphX - http://www.MorphoGraphX.org
// Copyright (C) 2012-2015 Richard S. Smith and collaborators.
//
// If you use MorphoGraphX in your work, please cite:
//   http://dx.doi.org/10.7554/eLife.05864
//
// MorphoGraphX is free software, and is licensed under under the terms of the 
// GNU General (GPL) Public License version 2.0, http://www.gnu.org/licenses.
// 
#ifndef MESH_PROCESS_SELECTION_HPP
#define MESH_PROCESS_SELECTION_HPP

#include <Process.hpp>

namespace mgx {
namespace process {
///\addtogroup MeshProcess
///@{
/**
 * \class MeshSelectAll
 *
 * Select all the vertices in the current mesh.
 */
class mgxBase_EXPORT MeshSelectAll : public MeshProcess {
public:
  MeshSelectAll(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select All";
  }
  QString description() const {
    return "Select all vertices of the current mesh";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectBadNormals
 *
 * Select all the vertices with invalid normals in the current mesh
 */
class mgxBase_EXPORT MeshSelectBadNormals : public MeshProcess {
public:
  MeshSelectBadNormals(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select Bad Normals";
  }
  QString description() const
  {
    return "Select all vertices of the current mesh with bad normals (i.e. normals is not opf size 1)";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshUnselect
 *
 * Ensure no vertex is selected in the current mesh.
 */
class mgxBase_EXPORT MeshUnselect : public MeshProcess {
public:
  MeshUnselect(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Unselect";
  }
  QString description() const {
    return "Unselect the vertices of the current mesh";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshInvertSelection
 *
 * Invert the selection status of all the vertices in the current mesh.
 */
class mgxBase_EXPORT MeshInvertSelection : public MeshProcess {
public:
  MeshInvertSelection(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Invert Selection";
  }
  QString description() const {
    return "Invert the selection of the current mesh.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectUnlabeled
 *
 * Select all the unlabel vertices of the current mesh. Other vertices may either be unselected, or stay in their
 * current selection state.
 */
class mgxBase_EXPORT MeshSelectUnlabeled : public MeshProcess {
public:
  MeshSelectUnlabeled(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    bool replace = stringToBool(parms[0]);
    return (*this)(currentMesh(), replace);
  }

  bool operator()(Mesh* m, bool replace);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select Unlabeled";
  }
  QString description() const {
    return "Add to or replace the selection with the unlabeled vertices.";
  }
  QStringList parmNames() const {
    return QStringList() << "Replace selection";
  }
  QStringList parmDescs() const {
    return QStringList() << "Replace selection";
  }
  QStringList parmDefaults() const {
    return QStringList() << "No";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = booleanChoice();
    return map;
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectLabeled
 *
 * Select all labeled vertices in the current mesh.
 */
class mgxBase_EXPORT MeshSelectLabeled : public MeshProcess {
public:
  MeshSelectLabeled(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    bool replace = stringToBool(parms[0]);
    return (*this)(currentMesh(), replace);
  }

  bool operator()(Mesh* m, bool replace);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select Labeled";
  }
  QString description() const {
    return "Add to or replace the selection with the labeled vertices.";
  }
  QStringList parmNames() const {
    return QStringList() << "Replace selection";
  }
  QStringList parmDescs() const {
    return QStringList() << "Replace selection";
  }
  QStringList parmDefaults() const {
    return QStringList() << "No";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = booleanChoice();
    return map;
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectLabel
 */
class mgxBase_EXPORT MeshSelectLabel : public MeshProcess {
public:
  MeshSelectLabel(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    bool replace = stringToBool(parms[0]);
    return (*this)(currentMesh(), replace, parms[1].toInt());
  }

  bool operator()(Mesh* m, bool replace, int label);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select Label";
  }
  QString description() const
  {
    return "Add to or replace the selection with the vertices of a given label (0 for current label).";
  }
  QStringList parmNames() const
  {
    return QStringList() << "Replace selection"
                         << "Label (0 for current)";
  }
  QStringList parmDefaults() const
  {
    return QStringList() << "No"
                         << "0";
  }
  ParmChoiceMap parmChoice() const
  {
    ParmChoiceMap map;
    map[0] = booleanChoice();
    return map;
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshUnselectLabel MeshProcessSelection.hpp <MeshProcessSelection.hpp>
 *
 * Unselect all the vertices having a given label.
 */
class mgxBase_EXPORT MeshUnselectLabel : public MeshProcess {
public:
  MeshUnselectLabel(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& parms)
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh(), parms[0].toInt());
  }

  bool operator()(Mesh* m, int label);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Unselect Label";
  }
  QString description() const
  {
    return "Remove the vertices of a given label (0 for current label) from the selection.";
  }
  QStringList parmNames() const {
    return QStringList() << "Label (0 for current)";
  }
  QStringList parmDescs() const {
    return QStringList() << "Label (0 for current)";
  }
  QStringList parmDefaults() const {
    return QStringList() << "0";
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectClip MeshProcessSelection.hpp <MeshProcessSelection.hpp>
 *
 * Select all vertices within the clipped region.
 */
class mgxBase_EXPORT MeshSelectClip : public MeshProcess {
public:
  MeshSelectClip(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select Clip Region";
  }
  QString description() const {
    return "Add vertices in clip region to selection.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectWholeLabelExtend MeshProcessSelection.hpp <MeshProcessSelection.hpp>
 *
 * Extent the current selection so each label having at least one vertex
 * selected will be fully selected.
 */
class mgxBase_EXPORT MeshSelectWholeLabelExtend : public MeshProcess {
public:
  MeshSelectWholeLabelExtend(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Extend to Whole Cells";
  }
  QString description() const {
    return "Extend Selection to Whole Cells";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class MeshSelectDuplicateCells MeshProcessSelection.hpp <MeshProcessSelection.hpp>
 *
 * Select vertices if the region with their label is not contiguous (e.g. will be seen as more than one cell).
 */
class mgxBase_EXPORT MeshSelectDuplicateCells : public MeshProcess {
public:
  MeshSelectDuplicateCells(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY | MESH_SHOW_MESH))
      return false;
    return (*this)(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Select Duplicate Cells";
  }
  QString description() const {
    return "Select cells with duplicate labels.";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon();
  }
};

/**
 * \class ExtendByConnectivity MeshProcessSelection.hpp <MeshProcessSelection.hpp>
 *
 * Extend the current selection to all vertices that are connected to a
 * currently selected vertex.
 */
class ExtendByConnectivity : public MeshProcess {
public:
  ExtendByConnectivity(const MeshProcess& process)
    : Process(process)
    , MeshProcess(process)
  {
  }

  bool operator()(const QStringList& )
  {
    if(!checkState().mesh(MESH_NON_EMPTY))
      return false;
    return operator()(currentMesh());
  }

  bool operator()(Mesh* m);

  QString folder() const {
    return "Selection";
  }
  QString name() const {
    return "Extend by Connectivity";
  }
  QString description() const {
    return "Extend the selection to connected regions";
  }
  QStringList parmNames() const {
    return QStringList();
  }
  QStringList parmDescs() const {
    return QStringList();
  }
  QIcon icon() const {
    return QIcon(":/images/SelectConnected.png");
  }
};
///@}
} // namespace process
} // namespace mgx

#endif
